from abc import ABC, abstractmethod
import inspect


class QueryVariable:
    @abstractmethod
    async def to_variable(self):
        raise NotImplementedError


async def parse_variables(variables):
    print(variables)
    parsed_kwargs = {}

    for key, value in variables.items():
        if hasattr(value, "to_variable"):
            print("nanan")
            parsed_kwargs[key] = await value.to_variable()
        else:
            parsed_kwargs[key] = value

    return parsed_kwargs



from abc import abstractproperty
from herre.exceptions import TokenExpired
from herre.wards.base import BaseQuery, BaseWard, WardException
import aiohttp
import logging
import re
from herre.wards.variables import parse_variables
from fakts import Fakts, get_current_fakts, Config

logger = logging.getLogger(__name__)

gqlparsed_with_variables = re.compile(
    r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)\((?P<arguments>[^\)]*)\)[\s]*{[\s]*(?P<firstchild>[^\(:]*).*"
)
gqlparser_without_variables = re.compile(
    r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)[\s]*{[\s]*(?P<firstchild>[^\(\{\s:]*).*"
)


class GQLException(Exception):
    pass


class ParsedQuery(BaseQuery):
    def __init__(self, query: str) -> None:
        self.query = query
        self.variables = None
        self._type = None
        self.m = gqlparsed_with_variables.match(self.query)
        self.has_variables = True

        if not self.m:
            self.m = gqlparser_without_variables.match(self.query)
            self.has_variables = False

        if not self.m:
            raise GQLException(f"Illformed request {self.query}")

    def get_operation_name_for_datakey(self, key):
        # TODO: Make work
        return key

    def combine(self, variables: dict):
        self.variables = variables
        return {"query": self.query, "variables": self.variables}

    def parsed(self):
        assert (
            self.variables and self.query
        ), "Please specify query and set variables before parsing"
        return {"query": self.query, "variables": self.variables}

    @property
    def firstchild(self):
        return self.m.group("firstchild")

    @property
    def operation_name(self):
        return self.m.group("operation")

    @property
    def type(self):
        return self.m.group("type")

    def extract(self, result: dict):
        assert (
            self.firstchild in result
        ), f"Cannot Access {self.firstchild}: in {result}"
        assert result[self.firstchild] is not None, f"Empty response {result}"
        return result[self.firstchild]


class GraphQLException(WardException):
    pass


class GraphQLQueryException(GraphQLException):
    pass


class GraphQLProtocolException(WardException):
    pass


class GraphQLWardConfig(Config):
    @abstractproperty
    def endpoint(self):
        return NotImplementedError("Please overwrite endpoint")


class GraphQLWard(BaseWard):
    configClass = GraphQLWardConfig
    config: GraphQLWardConfig

    class Meta:
        abstract = True

    async def handle_connect(self):

        self.async_session = aiohttp.ClientSession(headers=self.herre.headers)

    async def handle_disconnect(self):
        await self.async_session.close()
        self.connected = False

    async def handle_run(self, gql: ParsedQuery, variables: dict = {}, retry=0):
        print(variables)
        try:
            assert (
                retry < self.max_retries
            ), f"Retries Exceeded for Request {gql} with {variables}"
            async with self.async_session.post(
                self.config.endpoint, json={"query": gql.query, "variables": variables}
            ) as resp:

                if resp.status == 200:
                    result = await resp.json()
                    logger.debug(f"Received Reply {result}")

                    if "errors" in result:
                        raise GraphQLQueryException(
                            f"Ward {self.config.endpoint}:" + str(result["errors"])
                        )

                    return result["data"]

                if resp.status == 403:
                    logger.error("Auth token is expired trying to refresh")
                    raise TokenExpired("Token Expired Error")

                if resp.status == 400:
                    raise GraphQLProtocolException(await resp.json())

                raise GraphQLProtocolException(
                    f"Unexpected statuscode {resp.status} {resp}"
                )

        except TokenExpired as e:
            await self.herre.arefresh()
            return await self.run(gql, variables, retry=retry + 1)

        except aiohttp.client_exceptions.ClientOSError as e:
            await self.adisconnect()
            raise GraphQLProtocolException("Ward is not reachable") from e






from abc import abstractmethod

from pydantic.main import prepare_config
from herre.console.context import get_current_console
from herre.config import HerreConfig
from abc import ABC
import os 
import asyncio
import requests
import logging
import aiohttp
import shelve

from fakts import Fakts, get_current_fakts, Config

logger = logging.getLogger(__name__)


class GrantException(Exception):
    pass

class RetryException(GrantException):
    pass




class BaseGrant(ABC):
    refreshable = False
    type = None

    def __init__(self, herre_config: HerreConfig,  token_url="o/token/", authorize_url="o/authorize/", token_file: str = "token.temp", save_token = True, max_retries = 3, facts: Fakts = None, **kwargs) -> None:
        self.facts = facts or get_current_fakts()
        self.config = herre_config
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if self.config.secure else "1"


        self.base_url = f'{"https" if self.config.secure else "http"}://{self.config.host}:{self.config.port}{self.config.subpath}/'
        self.auth_url = self.base_url + authorize_url
        self.token_url = self.base_url + token_url
        self.refresh_url = self.token_url
        self._userinfo = None
       
        self.scopes = self.config.scopes + ["introspection"]
        self.scope = " ".join(self.scopes)
        self.token = None
        self._user = None
        self.token_file = token_file
        self.save_token = save_token
        self.max_retries = max_retries

        # State changed through Class
        self.can_refresh = self.refreshable


    @abstractmethod
    async def afetch_token(self, **kwargs):
        raise NotImplementedError()


    async def arefresh(self, **kwargs):
        if not self.refreshable: return None
        assert "refresh_token" in self.token, "Token had not refresh-token attached"
        async with aiohttp.ClientSession() as session:
            async with session.post(self.refresh_url, data={"grant_type": "refresh_token", "refresh_token": self.token["refresh_token"], "client_id": self.config.client_id, "client_secret": self.config.client_secret}) as resp:
                self.token =  await resp.json()
                assert "access_token" in self.token, "Returned refreshed token does not have an access_token"
                if "refresh_token" in self.token:
                    self.can_refresh = True

